@@ -19,6 +19,7 @@ from maitai._inference import (
 )
 from maitai._types import (
     Body,
+    EvaluateCallback,
     Headers,
     MaitaiChunk,
     MaitaiCompletion,
@@ -26,21 +27,14 @@ from maitai._types import (
     ToolFunction,
 )
 from maitai._utils import (
+    OpenAISchemaGenerator,
     convert_open_ai_chat_completion_chunk,
     convert_openai_chat_completion,
     get_chat_completion_params,
     required_args,
 )
 from maitai.tools import Tools
-from maitai_common.utils.proto_utils import openai_messages_to_proto
-from maitai_common.utils.types import EvaluateCallback
-from maitai_gen.chat import (
-    ChatCompletionParams,
-    ChatCompletionRequest,
-    ChatMessage,
-    ClientParams,
-    EvaluationContentType,
-)
+from maitai_gen.chat import ChatCompletionParams, ClientParams, EvaluationContentType
 from maitai_gen.config import InferenceLocations
 from maitai_gen.inference import InferenceStreamResponse
 from maitai_gen.metric import RequestTimingMetric
@@ -72,6 +66,10 @@ class Maitai:
             maitai_api_key = os.environ.get("MAITAI_API_KEY")
         if maitai_api_key:
             config.initialize(maitai_api_key)
+        else:
+            raise ValueError(
+                "MAITAI_API_KEY has not been set. Either set via environment variable or pass it directly into the client."
+            )
         if api_key:
             if not openai_api_key:
                 openai_api_key = api_key
@@ -262,7 +260,7 @@ class Completions:
             max_tokens = maitai_config.max_tokens
         if max_completion_tokens and max_completion_tokens != openai.NOT_GIVEN:
             max_tokens = max_completion_tokens
-        if n == openai.NOT_GIVEN:
+        if n == openai.NOT_GIVEN and maitai_config.n:
             n = maitai_config.n
         if frequency_penalty == openai.NOT_GIVEN:
             frequency_penalty = maitai_config.frequency_penalty
@@ -295,37 +293,39 @@ class Completions:
             top_p=top_p,
             user=user,
             parallel_tool_calls=parallel_tool_calls,
+            extra_headers=extra_headers,
         )
         if server_side_inference:
             response_timeout = None
             if isinstance(timeout, float) or isinstance(timeout, int):
                 response_timeout = timeout
-            chat_request: ChatCompletionRequest = ChatCompletionRequest(
-                application_ref_name=application_ref_name,
-                session_id=session_id,
-                reference_id=reference_id,
-                action_type=intent,
-                apply_corrections=apply_corrections,
-                evaluation_enabled=evaluation_enabled,
-                params=completion_params,
-                return_evaluation=True if callback else False,
-                context_retrieval_enabled=context_retrieval_enabled,
-                context_query=context_query,
-                return_request=return_request,
-                fallback_model=fallback_model,
-                user_id=user_id,
-                assistant=assistant,
-                auth_keys=config.auth_keys,
-                metadata=metadata,
-                safe_mode=safe_mode,
-                extra_headers=extra_headers or {},
-            )
+            request_dict = {
+                "application_ref_name": application_ref_name,
+                "session_id": session_id,
+                "reference_id": reference_id,
+                "action_type": intent,
+                "apply_corrections": apply_corrections,
+                "evaluation_enabled": evaluation_enabled,
+                "params": completion_params,
+                "return_evaluation": True if callback else False,
+                "context_retrieval_enabled": context_retrieval_enabled,
+                "context_query": context_query,
+                "return_request": return_request,
+                "fallback_model": fallback_model,
+                "user_id": user_id,
+                "assistant": assistant,
+                "auth_keys": config.auth_keys.to_dict(),
+                "metadata": metadata,
+                "safe_mode": safe_mode,
+            }
+
             if self._client_params is not None:
-                chat_request.client_params = self._client_params
+                request_dict["client_params"] = self._client_params.to_dict()
+
             try:
                 timing.time_pre_request = time.time()
                 response = maitai.Inference.infer(
-                    chat_request, callback, response_timeout
+                    request_dict, callback, response_timeout
                 )
                 if stream:
                     if "self" in retry_args:
@@ -355,13 +355,17 @@ class Completions:
                     timing.request_id = resp.chat_completion_response.request_id
                     timing.time_request_end = time.time()
                     maitai.Inference.store_request_timing_data(timing)
-                    return MaitaiCompletion(resp.chat_completion_response)
+                    return MaitaiCompletion(
+                        resp.chat_completion_response, response_format=response_format
+                    )
                 raise InferenceException("No data received")
             except MaitaiConnectionError as e:
                 if "gpt" not in model and fallback_model and "gpt" in fallback_model:
                     model = fallback_model
                 if "gpt" not in model:
                     raise e
+                if safe_mode:
+                    raise e
                 print("Maitai issue, retrying on client side")
                 server_side_inference = False
         if not server_side_inference:
@@ -371,6 +375,13 @@ class Completions:
                     default_headers=self._client_params.default_headers or None,
                     base_url=self._client_params.base_url or None,
                 )
+            if hasattr(response_format, "model_json_schema"):
+                response_format = {
+                    "type": "json_schema",
+                    "json_schema": response_format.model_json_schema(
+                        schema_generator=OpenAISchemaGenerator, mode="serialization"
+                    ),
+                }
             response = self._client.chat.completions.create(
                 messages=messages,
                 model=model,
@@ -416,20 +427,19 @@ class Completions:
             else:
                 timing.time_request_end = time.time()
                 maitai_completion = convert_openai_chat_completion(response)
-                proto_messages = openai_messages_to_proto(messages)
-                proto_messages.append(
-                    ChatMessage(
-                        role="assistant",
-                        content=maitai_completion.choices[0].message.content,
-                    )
-                )
+                full_conv = messages + [
+                    {
+                        "role": "assistant",
+                        "content": maitai_completion.choices[0].message.content,
+                    }
+                ]
                 if evaluation_enabled:
                     maitai.Evaluator.evaluate(
                         session_id=session_id,
                         reference_id=reference_id,
                         intent=intent,
                         content_type=EvaluationContentType.MESSAGE,
-                        content=proto_messages,
+                        content=full_conv,
                         application_ref_name=application_ref_name,
                         callback=callback,
                         chat_completion_response=maitai_completion,
@@ -450,7 +460,9 @@ class Completions:
                         timing=timing,
                         metadata=metadata,
                     )
-                return MaitaiCompletion(maitai_completion)
+                return MaitaiCompletion(
+                    maitai_completion, response_format=response_format
+                )
 
 
 def _process_openai_stream(
@@ -467,7 +479,6 @@ def _process_openai_stream(
     metadata: dict = {},
 ) -> Iterable[MaitaiChunk]:
     full_body = ""
-    proto_messages = openai_messages_to_proto(messages)
     last_chunk = None
     for chunk in stream:
         if last_chunk is None:
@@ -483,13 +494,13 @@ def _process_openai_stream(
         yield MaitaiChunk(maitai_chunk)
     timing.time_request_end = time.time()
     if evaluation_enabled:
-        proto_messages.append(ChatMessage(role="assistant", content=full_body))
+        full_conv = messages + [{"role": "assistant", "content": full_body}]
         Evaluator.evaluate(
             session_id=session_id,
             reference_id=reference_id,
             intent=intent,
             content_type=EvaluationContentType.MESSAGE,
-            content=proto_messages,
+            content=full_conv,
             application_ref_name=application_ref_name,
             callback=callback,
             chat_completion_chunk=last_chunk,
