@@ -62,8 +62,15 @@ class MaitaiAsync:
         http_client: Optional[httpx.Client] = None,
         _strict_response_validation: bool = False,
     ):
+        if not maitai_api_key:
+            maitai_api_key = os.environ.get("MAITAI_API_KEY")
         if maitai_api_key:
             config.initialize(maitai_api_key)
+        else:
+            raise ValueError(
+                "MAITAI_API_KEY has not been set. Either set via environment variable or pass it directly into the client."
+            )
+
         if api_key:
             if not openai_api_key:
                 openai_api_key = api_key
@@ -254,7 +261,7 @@ class AsyncCompletions:
             max_tokens = maitai_config.max_tokens
         if max_completion_tokens and max_completion_tokens != openai.NOT_GIVEN:
             max_tokens = max_completion_tokens
-        if n == openai.NOT_GIVEN:
+        if n == openai.NOT_GIVEN and maitai_config.n:
             n = maitai_config.n
         if frequency_penalty == openai.NOT_GIVEN:
             frequency_penalty = maitai_config.frequency_penalty
@@ -357,6 +364,8 @@ class AsyncCompletions:
                     model = fallback_model
                 if "gpt" not in model:
                     raise e
+                if safe_mode:
+                    raise e
                 print("Maitai issue, retrying on client side")
                 server_side_inference = False
         if not server_side_inference:
